;;; -*- syntax: common-lisp; package: cmn; base: 10; mode: lisp -*-
;;;
;;; Postscript, X, Braille, or Quickdraw output
;;;
;;; massively rewritten in Nov-97: CMN outside of this file acts as though 
;;; it were driving a relatively dumb Postscript interpreter, and we provide
;;; here the necessary matrix/line-width/polygon/Bezier/font stuff to make
;;; that work in Quickdraw or X, etc.

;;; finish Braille, add PDF from cl-pdf

(in-package :cmn)

;;; qddirect option added 5-June-97 courtesy Gerard Assayag
;;; (on explicit #+mcl switch in some cases to avoid package confusion in other lisps)
;;; and I (bil) moved some code to cmn-mcl.lisp 
;;;   (e.g. in ACL the #_ stuff gets the error "No dispatch function defined for #\_." even when it's under #+mcl)
;;;
;;; to port to a new graphics environment, we need at a minimum lineto, moveto, and a polygon filler.
;;; next step involves text, arc, font/color/line-width settings.
;;; then curveto (Bezier curves), dashed lines, various filling routines, etc.
;;;
;;; Postscript's fill routine is substantially smarter than either X's or Quickdraw's.
;;; In some cases, we can get the latter to work by using the "even-odd" rule, rather than
;;; the default "winding" rule, modulo screwing around with curve direction and so on.
;;; A number of music symbols are complex enough that nothing works completely -- I'll
;;; probably have to make two versions, one broken into separate pieces for X and Quickdraw.



#-(or mcl lispworks) (defmacro make-short-float (x) `(coerce ,x 'short-float))
#+lispworks (defmacro make-short-float (x) `(coerce ,x 'single-float))
#+mcl (defmacro make-short-float (x) `(coerce ,x 'float)) ;not the same as (float x)!!
#-lispworks (defvar c-args (make-array 6 :element-type #-mcl 'short-float #+mcl 'float :initial-element (coerce 0.0 #-mcl 'short-float #+mcl 'float)))
#+lispworks (defvar c-args (sys:in-static-area (make-array 6 :element-type 'single-float :initial-element (coerce 0.0 'single-float))))



;;; -------------------------------- GRAPHICS CONTEXT VARIABLES --------------------------------

(defvar *cmn-top-y* 0)			; for QuickDraw and X y axis inversion (both assume (0,0) is top left corner)
					; this is set in cmn4.lisp when we know where the top left corner is in Postscript coordinates

(defvar grf-m0 1.0)
(defvar grf-m1 0.0)
(defvar grf-m2 0.0)
(defvar grf-m3 1.0)
(defvar grf-xoff 0.0)
(defvar grf-yoff 0.0)
(defvar grf-xp 0.0)
(defvar grf-yp 0.0)
(defvar grf-xpoints nil)
(defvar grf-width 0)
(defvar grf-color nil)
(defvar grf-font (normal-font))
(defvar grf-fontsize 40)
(defvar grf-active-font nil)
(defvar grf-active-fontsize 0)
(defvar grf-matrix-stack nil)
(defvar grf-save-stack nil)
(defvar xcmnw-open nil)

(defvar *grf-curve-accuracy* 5)


;;; -------------------------------- FOREIGN FUNCTIONS --------------------------------

;;; sgicmn.c
#+(and excl cltl2 (not allegro-v5.0) (not allegro-v6.0) (not allegro-v6.1) (not allegro-v6.2) (not allegro-v7.0) (or sgi sun linux) (not only-lisp))
  (progn
    (ff:defforeign 'cmn-popen :entry-point "cmn_popen" :arguments '(string) :return-type :void)
    (ff:defforeign 'cmn-pclose :entry-point "cmn_pclose" :arguments nil :return-type :integer)
    (ff:defforeign 'x-clear-window :entry-point "x_clear_window" :arguments nil :return-type :void)
    (ff:defforeign 'x-lineto :entry-point "x_lineto" :arguments '(array) :return-type :void)
    (ff:defforeign 'x-polygon :entry-point "x_polygon" :arguments '(integer integer integer array) :return-type :void)
    (ff:defforeign 'x-rectangle :entry-point "x_rectangle" :arguments '(integer array) :return-type :void)
    (ff:defforeign 'x-circle :entry-point "x_circle" :arguments '(integer array) :return-type :void)
    (ff:defforeign 'x-text :entry-point "x_text" :arguments '(integer array string string) :return-type :void)
    (ff:defforeign 'x-color :entry-point "x_color" :arguments '(integer integer integer) :return-type :void)
    (ff:defforeign 'x-line-width :entry-point "x_line_width" :arguments '(integer) :return-type :void)
    (ff:defforeign 'x-fill-style :entry-point "x_fill_style" :arguments '(integer) :return-type :void)
    (ff:defforeign 'x-line-style :entry-point "x_line_style" :arguments '(integer) :return-type :void))

#+(and allegro-v5.0 (or sgi sun linux) (not only-lisp))
  (progn
    (ff:def-foreign-call (cmn-popen "cmn_popen") ((file (* :char) string)) :returning :void)
    (ff:def-foreign-call (cmn-pclose "cmn_pclose") (:void) :returning :int)
    (ff:def-foreign-call (x-clear-window "x_clear_window") (:void) :returning :void)
    (ff:def-foreign-call (x-lineto "x_lineto") ((vals (* :int) array)) :returning :void)
    (ff:def-foreign-call (x-polygon "x_polygon") ((filled :int) (eo :int) (size :int) (vals (* :int) array)) :returning :void)
    (ff:def-foreign-call (x-rectangle "x_rectangle") ((filled :int) (vals (* :int) array)) :returning :void)
    (ff:def-foreign-call (x-circle "x_circle") ((filled :int) (vals (* :int) array)) :returning :void)
    (ff:def-foreign-call (x-text "x_text") ((size :int) (vals (* :int) array) (font (* :char) string) (text (* :char) string)) :returning :void)
    (ff:def-foreign-call (x-color "x_color") ((r :int) (g :int) (b :int)) :returning :void)
    (ff:def-foreign-call (x-line-width "x_line_width") ((r :int)) :returning :void)
    (ff:def-foreign-call (x-fill-style "x_fill_style") ((r :int)) :returning :void)
    (ff:def-foreign-call (x-line-style "x_line_style") ((r :int)) :returning :void)
    #+linux (ff:def-foreign-call (cmn-send-snd "cmn_send_snd") ((cmd (* :char) string)) :returning :void)
    #+linux (ff:def-foreign-call (cmn-init-x "cmn_init_x") (:void) :returning :void)
    #+linux (ff:def-foreign-call (cmn-receive-snd-2 "cmn_receive_snd") (:void) :returning :int)
    )
#+(and excl linux allegro-v5.0 (not only-lisp))
  (defun cmn-receive-snd-1 ()
    (let ((ptr (cmn-receive-snd-2)))
      (if (not (zerop ptr))
	  (ff:char*-to-string ptr))))

#+(and excl allegro-v6.0 (not acl-61) (or sgi sun linux) (not only-lisp))
  (progn
    (ff:def-foreign-call (cmn-popen "cmn_popen") ((file (* :char))) :returning :void)
    (ff:def-foreign-call (cmn-pclose "cmn_pclose") (:void) :returning :int)
    (ff:def-foreign-call (x-clear-window "x_clear_window") (:void) :returning :void)
    (ff:def-foreign-call (x-lineto "x_lineto") ((vals (* :int) array)) :returning :void)
    (ff:def-foreign-call (x-polygon "x_polygon") ((filled :int) (eo :int) (size :int) (vals (* :int) array)) :returning :void)
    (ff:def-foreign-call (x-rectangle "x_rectangle") ((filled :int) (vals (* :int) array)) :returning :void)
    (ff:def-foreign-call (x-circle "x_circle") ((filled :int) (vals (* :int) array)) :returning :void)
    (ff:def-foreign-call (x-text "x_text") ((size :int) (vals (* :int) array) (font (* :char)) (text (* :char))) :returning :void)
    (ff:def-foreign-call (x-color "x_color") ((r :int) (g :int) (b :int)) :returning :void)
    (ff:def-foreign-call (x-line-width "x_line_width") ((r :int)) :returning :void)
    (ff:def-foreign-call (x-fill-style "x_fill_style") ((r :int)) :returning :void)
    (ff:def-foreign-call (x-line-style "x_line_style") ((r :int)) :returning :void)
    #+linux (ff:def-foreign-call (cmn-send-snd "cmn_send_snd") ((cmd (* :char))) :returning :void)
    #+linux (ff:def-foreign-call (cmn-init-x "cmn_init_x") (:void) :returning :void)
    #+linux (ff:def-foreign-call (cmn-receive-snd-1 "cmn_receive_snd") (:void) :returning ((* :char)))
    )

#+(and excl allegro-v6.0 acl-61 (or sgi sun linux) (not only-lisp))
  (progn
    (ff:def-foreign-call (cmn-popen "cmn_popen")
			 ((file (* :char)))
			 :strings-convert t
			 :returning :void)
    (ff:def-foreign-call (cmn-pclose "cmn_pclose") (:void) :returning :int)
    (ff:def-foreign-call (x-clear-window "x_clear_window") (:void) :returning :void)
    (ff:def-foreign-call (x-lineto "x_lineto") ((vals (* :int) array)) :returning :void)
    (ff:def-foreign-call (x-polygon "x_polygon") ((filled :int) (eo :int) (size :int) (vals (* :int) array)) :returning :void)
    (ff:def-foreign-call (x-rectangle "x_rectangle") ((filled :int) (vals (* :int) array)) :returning :void)
    (ff:def-foreign-call (x-circle "x_circle") ((filled :int) (vals (* :int) array)) :returning :void)
    (ff:def-foreign-call (x-text "x_text")
			 ((size :int) (vals (* :int) array) (font (* :char)) (text (* :char)))
			 :strings-convert t
			 :returning :void)
    (ff:def-foreign-call (x-color "x_color") ((r :int) (g :int) (b :int)) :returning :void)
    (ff:def-foreign-call (x-line-width "x_line_width") ((r :int)) :returning :void)
    (ff:def-foreign-call (x-fill-style "x_fill_style") ((r :int)) :returning :void)
    (ff:def-foreign-call (x-line-style "x_line_style") ((r :int)) :returning :void)
    #+linux (ff:def-foreign-call (cmn-send-snd "cmn_send_snd")
				 ((cmd (* :char)))
				 :strings-convert t
				 :returning :void)
    #+linux (ff:def-foreign-call (cmn-init-x "cmn_init_x") (:void) :returning :void)
    #+linux (ff:def-foreign-call (cmn-receive-snd-1 "cmn_receive_snd")
				 (:void)
				 :strings-convert t
				 :returning ((* :char)))
    )

;;; this is becoming absurd -- it speeds up CMN by maybe a factor of 2, but is a headache in every other regard
#+(and excl allegro-v7.0 (or sgi sun linux) (not only-lisp))
  (progn
    (ff:def-foreign-call (cmn-popen "cmn_popen")
			 ((file (* :char)))
			 :strings-convert t
			 :returning :void)
    (ff:def-foreign-call (cmn-pclose "cmn_pclose") (:void) :returning :int)
    (ff:def-foreign-call (x-clear-window "x_clear_window") (:void) :returning :void)
    (ff:def-foreign-call (x-lineto "x_lineto") ((vals (:array :int))) :returning :void)
    (ff:def-foreign-call (x-polygon "x_polygon") ((filled :int) (eo :int) (size :int) (vals (:array :int))) :returning :void)
    (ff:def-foreign-call (x-rectangle "x_rectangle") ((filled :int) (vals (:array :int))) :returning :void)
    (ff:def-foreign-call (x-circle "x_circle") ((filled :int) (vals (:array :int))) :returning :void)
    (ff:def-foreign-call (x-text "x_text")
			 ((size :int) (vals (:array :int)) (font (* :char)) (text (* :char)))
			 :strings-convert t
			 :returning :void)
    (ff:def-foreign-call (x-color "x_color") ((r :int) (g :int) (b :int)) :returning :void)
    (ff:def-foreign-call (x-line-width "x_line_width") ((r :int)) :returning :void)
    (ff:def-foreign-call (x-fill-style "x_fill_style") ((r :int)) :returning :void)
    (ff:def-foreign-call (x-line-style "x_line_style") ((r :int)) :returning :void)
    #+linux (ff:def-foreign-call (cmn-send-snd "cmn_send_snd")
				 ((cmd (* :char)))
				 :strings-convert t
				 :returning :void)
    #+linux (ff:def-foreign-call (cmn-init-x "cmn_init_x") (:void) :returning :void)
    #+linux (ff:def-foreign-call (cmn-receive-snd-1 "cmn_receive_snd")
				 (:void)
				 :strings-convert t
				 :returning ((* :char)))
    )



#+(and cmu (not only-lisp)) (use-package "ALIEN")
;#+cmu (use-package "C-CALL")
#+(and cmu (not only-lisp))
(defun array-data-address (array)
       "Return the physical address of where the actual data of an array is
     stored.

     ARRAY must be a specialized array type in CMU Lisp.  This means ARRAY
     must be an array of one of the following types:

                       double-float
                       single-float
                       (unsigned-byte 32)
                       (unsigned-byte 16)
                       (unsigned-byte  8)
                       (signed-byte 32)
                       (signed-byte 16)
                       (signed-byte  8)
     "
       (declare (type (or #+signed-array (array (signed-byte 8))
                          #+signed-array (array (signed-byte 16))
                          #+signed-array (array (signed-byte 32))
                          (array (unsigned-byte 8))
                          (array (unsigned-byte 16))
                          (array (unsigned-byte 32))
                          (array single-float)
                          (array double-float))
                      array)
                (optimize (speed 3) (safety 0))
                (ext:optimize-interface (safety 3)))
       ;; with-array-data will get us to the actual data.  However, because
       ;; the array could have been displaced, we need to know where the
       ;; data starts.
       (lisp::with-array-data ((data array)
                               (start)
                               (end))
         (declare (ignore end))
         ;; DATA is a specialized simple-array.  Memory is laid out like this:
         ;;
         ;;   byte offset    Value
         ;;        0         type code (should be 70 for double-float vector)
         ;;        4         4 * number of elements in vector
         ;;        8         1st element of vector
         ;;      ...         ...
         ;;
         (let ((addr (+ 8 (logandc1 7 (kernel:get-lisp-obj-address data))))
               (type-size (let ((type (array-element-type data)))
                            (cond ((or (equal type '(signed-byte 8))
                                       (equal type '(unsigned-byte 8)))
                                   1)
                                  ((or (equal type '(signed-byte 16))
                                       (equal type '(unsigned-byte 16)))
                                   2)
                                  ((or (equal type '(signed-byte 32))
                                       (equal type '(unsigned-byte 32)))
                                   4)
                                  ((equal type 'single-float)
                                   4)
                                  ((equal type 'double-float)
                                   8)
                                  (t
                                   (error "Unknown specialized array element type"))))))
           (declare (type (unsigned-byte 32) addr)
                    (optimize (speed 3) (safety 0) (ext:inhibit-warnings 3)))
           (system:int-sap (the (unsigned-byte 32)
                             (+ addr (* type-size start)))))))

#+(and cmu (or linux sun sgi) (not only-lisp))
  (progn
    (def-alien-routine ("cmn_popen" cmn-popen) c-call:void (file c-call:c-string))
    (def-alien-routine ("cmn_pclose" cmn-pclose) c-call:void)
    (def-alien-routine ("x_clear_window" x-clear-window) c-call:void)
    (def-alien-routine ("x_lineto" x-lineto-1) c-call:void (vals (* c-call:int)))
    (defun x-lineto (vals) (x-lineto-1 (array-data-address vals)))
    (def-alien-routine ("x_polygon" x-polygon-1) c-call:void (filled c-call:int) (eo c-call:int) (size c-call:int) (vals (* c-call:int)))
    (defun x-polygon (filled eo size vals) (x-polygon-1 filled eo size (array-data-address vals)))
    (def-alien-routine ("x_rectangle" x-rectangle-1) c-call:void (filled c-call:int) (vals (* c-call:int)))
    (defun x-rectangle (filled vals) (x-rectangle-1 filled (array-data-address vals)))
    (def-alien-routine ("x_circle" x-circle-1) c-call:void (filled c-call:int) (vals (* c-call:int)))
    (defun x-circle (filled vals) (x-circle-1 filled (array-data-address vals)))
    (def-alien-routine ("x_text" x-text-1) c-call:void (size c-call:int) (vals (* c-call:int)) (font c-call:c-string) (text c-call:c-string))
    (defun x-text (size vals font text) (x-text-1 size (array-data-address vals) font text))
    (def-alien-routine ("x_color" x-color) c-call:void (r c-call:int) (g c-call:int) (b c-call:int))
    (def-alien-routine ("x_line_width" x-line-width) c-call:void (r c-call:int))
    (def-alien-routine ("x_fill_style" x-fill-style) c-call:void (r c-call:int))
    (def-alien-routine ("x_line_style" x-line-style) c-call:void (r c-call:int))
    )

;; from contrib/sb-grovel/array-data.lisp
#+(and sbcl (not only-lisp))
(defun array-data-address (array)
	 "Return the physical address of where the actual data of an array is
stored.

ARRAY must be a specialized array type - an array of one of these types:

                  double-float
                  single-float
                  (unsigned-byte 32)
                  (unsigned-byte 16)
                  (unsigned-byte  8)
                  (signed-byte 32)
                  (signed-byte 16)
                  (signed-byte  8)
"
  (declare (type (or (array (signed-byte 8))
		     (array base-char)
		     simple-base-string
                     (array (signed-byte 16))
                     (array (signed-byte 32))
                     (array (unsigned-byte 8))
                     (array (unsigned-byte 16))
                     (array (unsigned-byte 32))
                     (array single-float)
                     (array double-float))
                 array)
           (optimize (speed 0) (debug 3) (safety 3)))
  ;; with-array-data will get us to the actual data.  However, because
  ;; the array could have been displaced, we need to know where the
  ;; data starts.

  (let* ((type (car (multiple-value-list (array-element-type array))))
	 (type-size
	  (cond ((or (equal type '(signed-byte 8))
		     (equal type 'cl::base-char)
		     (equal type '(unsigned-byte 8)))
		 1)
		((or (equal type '(signed-byte 16))
		     (equal type '(unsigned-byte 16)))
		 2)
		((or (equal type '(signed-byte 32))
		     (equal type '(unsigned-byte 32)))
		 4)
		((equal type 'single-float)
		 4)
		((equal type 'double-float)
		 8)
		(t (error "Unknown specialized array element type")))))
    (sb-kernel::with-array-data ((data array)
		      (start)
		      (end))
      (declare (ignore end))
      ;; DATA is a specialized simple-array.  Memory is laid out like this:
      ;;
      ;;   byte offset    Value
      ;;        0         type code (e.g. 70 for double-float vector)
      ;;        4         FIXNUMIZE(number of elements in vector)
      ;;        8         1st element of vector
      ;;      ...         ...
      ;;
      (let* ((addr (+ 8 (logandc1 7 (sb-kernel:get-lisp-obj-address data)))))
	(declare (type (unsigned-byte 32) addr)
		 (optimize (speed 3) (safety 0)))
	(sb-sys:int-sap (the (unsigned-byte 32)
			  (+ addr (* type-size start))))))))
#+(and sbcl (not only-lisp))
(progn
    (sb-alien:define-alien-routine ("cmn_popen" cmn-popen) sb-alien:void (file sb-alien:c-string))
    (sb-alien:define-alien-routine ("cmn_pclose" cmn-pclose) sb-alien:void)
    (sb-alien:define-alien-routine ("x_clear_window" x-clear-window) sb-alien:void)
    (sb-alien:define-alien-routine ("x_lineto" x-lineto-1) sb-alien:void (vals (* sb-alien:int)))
    (defun x-lineto (vals) (x-lineto-1 (array-data-address vals)))
    (sb-alien:define-alien-routine ("x_polygon" x-polygon-1) sb-alien:void
				   (filled sb-alien:int) (eo sb-alien:int) (size sb-alien:int) (vals (* sb-alien:int)))
    (defun x-polygon (filled eo size vals) (x-polygon-1 filled eo size (array-data-address vals)))
    (sb-alien:define-alien-routine ("x_rectangle" x-rectangle-1) sb-alien:void (filled sb-alien:int) (vals (* sb-alien:int)))
    (defun x-rectangle (filled vals) (x-rectangle-1 filled (array-data-address vals)))
    (sb-alien:define-alien-routine ("x_circle" x-circle-1) sb-alien:void (filled sb-alien:int) (vals (* sb-alien:int)))
    (defun x-circle (filled vals) (x-circle-1 filled (array-data-address vals)))
    (sb-alien:define-alien-routine ("x_text" x-text-1) sb-alien:void
				   (size sb-alien:int) (vals (* sb-alien:int)) (font sb-alien:c-string) (text sb-alien:c-string))
    (defun x-text (size vals font text) (x-text-1 size (array-data-address vals) font text))
    (sb-alien:define-alien-routine ("x_color" x-color) sb-alien:void (r sb-alien:int) (g sb-alien:int) (b sb-alien:int))
    (sb-alien:define-alien-routine ("x_line_width" x-line-width) sb-alien:void (r sb-alien:int))
    (sb-alien:define-alien-routine ("x_fill_style" x-fill-style) sb-alien:void (r sb-alien:int))
    (sb-alien:define-alien-routine ("x_line_style" x-line-style) sb-alien:void (r sb-alien:int))
    )

#+(and openmcl with-x11)
(progn
  (defmacro with-c-filename ((p f) &body body)
    `(ccl:with-cstrs ((,p (namestring (translate-logical-pathname ,f))))
		     ,@body))

  (defun heap-int* (hiarray &aux (p (ccl:%null-ptr)))
    (ccl::%vect-data-to-macptr hiarray p)
    p)

  (defun cmn-popen (file) (ccl:with-cstrs ((f file)) (ccl::external-call "_cmn_popen" :address f :void)))
  (defun cmn-pclose () (ccl::external-call "_cmn_pclose" :void))
  (defun x-clear-window () (ccl::external-call "_x_clear_window" :void))
  (defun x-lineto (vals) (ccl::external-call "_x_lineto" :address (heap-int* vals) :void))
  (defun x-polygon (filled eo size vals) (ccl::external-call "_x_polygon" :signed filled :signed eo :signed size :address (heap-int* vals) :void))
  (defun x-rectangle (filled vals) (ccl::external-call "_x_rectangle" :signed filled :address (heap-int* vals) :void))
  (defun x-circle (filled vals) (ccl::external-call "_x_circle" :signed filled :address (heap-int* vals) :void))
  (defun x-text (size vals font text)
    (ccl:with-cstrs ((f1 font))
      (ccl:with-cstrs ((f2 text))	       
        (ccl::external-call "_x_text" :signed size :address (heap-int* vals) :address f1 :address f2 :void))))
  (defun x-color (a b c) (ccl::external-call "_x_color" :signed a :signed b :signed c :void))
  (defun x-line-width (a) (ccl::external-call "_x_line_width" :signed a :void))
  (defun x-fill-style (a) (ccl::external-call "_x_fill_style" :signed a :void))
  (defun x-line-style (a) (ccl::external-call "_x_line_style" :signed a :void))
  )


#-openmcl
(defun cmn-receive-snd ()
  (let ((val (cmn-receive-snd-1)))
    (if (and val (stringp val) (> (length val) 0))
	(eval (read-from-string val)))))
#+openmcl (defun cmn-receive-snd () nil)


;; make the compiler happy
#+(or only-lisp (not (or (and openmcl with-x11) (and (or excl cmu sbcl) (or sgi sun linux)))))
  (progn
    (defun cmn-popen (str) (declare (ignore str)) )
    (defun cmn-pclose () )
    (defun x-clear-window () )
    (defun x-lineto (x) (declare (ignore x)) )
    (defun x-polygon (a b c d) (declare(ignore a b c d)) )
    (defun x-rectangle (a b) (declare (ignore a b)) )
    (defun x-circle (a b) (declare (ignore a b)) )
    (defun x-text (a b c d) (declare(ignore a b c d)) )
    (defun x-color (a b c) (declare(ignore a b c)) )
    (defun x-line-width (a) (declare (ignore a)) )
    (defun x-fill-style (a) (declare (ignore a)) )
    (defun x-line-style (a) (declare (ignore a)) ))

#-mcl
  (progn 
    (defun d_pensize (a b) (declare (ignore a b)) )
    (defun d_lineto (a b) (declare (ignore a b)) )
    (defun d_moveto (a b) (declare (ignore a b)) )
    (defun d_openpoly () )
    (defun d_closepoly () )
    (defun d_paintpoly (a) (declare (ignore a)) )
    (defun d_killpoly (a) (declare (ignore a)) )
    )


;;; -------------------------------- CLEAR GRAPHICS CONTEXT --------------------------------

(defun g-clear-context (score)
  (setf grf-m0 1.0)
  (setf grf-m1 0.0)
  (setf grf-m2 0.0)
  (setf grf-m3 1.0)
  (setf grf-xoff 0.0)
  (setf grf-yoff 0.0)
  (setf grf-xp 0.0)
  (setf grf-yp 0.0)
  (setf grf-xpoints nil)
  (setf grf-width 0)
  (setf grf-color nil)
  (setf grf-font (normal-font))
  (setf grf-fontsize (scr-size score))
  (setf grf-active-font nil)
  (setf grf-active-fontsize 0)
  (setf grf-matrix-stack nil)
  (setf grf-save-stack nil))



;;; -------------------------------- TRANSLATE POSITIONS --------------------------------
;;;
;;; both X and Quickdraw are upside down from Ps's point of view.  We also handle the
;;; coordinate transformations here.

(defun position-x (x y)
  (round (+ grf-xoff (* x grf-m0) (* y grf-m2))))

(defun position-y (x y)
  (round (- *cmn-top-y* (+ grf-yoff (* y grf-m3) (* x grf-m1)))))



;;; -------------------------------- TRANSLATE FONT NAMES --------------------------------

(defun Ps-to-X-font-name (name size)
  ;; "Times-Roman" 14 -> "-*-times-medium-*-*-*-14-*-*-*-*-*-*"
  (let* ((dashloc (position #\- name))
	 (family-name (if dashloc (subseq name 0 dashloc) name))
	 (type (if dashloc (subseq name (1+ dashloc)))))
    (concatenate 'string 
		 "-*-"
		 (string-downcase family-name) 
		 "-"
		 (if type
		     (if (or (string-equal type "roman") (string-equal type "medium"))
			 "medium-r-"
		       (if (string-equal type "italic")
			   "medium-i-"
			 (if (string-equal type "oblique")
			     "medium-o-"
			   (if (string-equal type "bold") 
			       "bold-r-"
			     (if (string-equal type "bolditalic")
				 "bold-i-"
			       "*-*-")))))
		   "*-*-")
		 "*-*-"
		 (format nil "~D" (or size 20))
		 "-*-*-*-*-*-*")))

(defun Ps-to-Qd-font-name (name)
  ;; we'll take the portion up to the "-" as the family name, and anything trailing as
  ;; the style name -- turned into a symbol
  (let* ((dashloc (position #\- name))
	 (family-name (if dashloc (subseq name 0 dashloc) name))
	 (type (if dashloc (subseq name (1+ dashloc)))))
    (list family-name
	  (if type
	      (if (or (string-equal type "italic") (string-equal type "oblique")) :italic
		(if (string-equal type "bold") :bold
		  ;; bolditalic occurs on NeXT
		  :plain))
	    :plain))))

(defun encode-text-face (n)		;Inside Mac I-210 -- these bits can obviously by OR'd
   (if (eq n :italic) 2 
     (if (eq n :bold) 1
       (if (eq n :outline) 8
	 (if (eq n :shadow) 16
	   0)))))


;;; -------------------------------- G-COMMENT --------------------------------

(defun g-comment (score x)
  ;; place a comment in the graphics output stream 
  ;; this is purely a debugging convenience, so it can be ignored
  (case (output-type score)
    (:postscript (format (output score) "% ~A~%" x))
    (:C	(c-comment x))
    (otherwise ))
  nil)


;;; -------------------------------- G-PUSH and G-POP (matrix handlers) --------------------------------

(defun g-push (score matrix)
  ;; matrix is the transformation matrix from the current coordinate system to the new one
  (case (output-type score)
    (:postscript (format (output score) " gsave [ ~{~A ~}] concat~%" (map 'list #'not-rational matrix)))
    (:C
     (setf (aref c-args 0) (make-short-float (first matrix)))
     (setf (aref c-args 1) (make-short-float (second matrix)))
     (setf (aref c-args 2) (make-short-float (third matrix)))
     (setf (aref c-args 3) (make-short-float (fourth matrix)))
     (setf (aref c-args 4) (make-short-float (fifth matrix)))
     (setf (aref c-args 5) (make-short-float (sixth matrix)))
     (c-matrix c-args))
    (otherwise ;do it by hand
     (let* ((old-matrix (list grf-m0 grf-m1 grf-m2 grf-m3 grf-xoff grf-yoff grf-xp grf-yp))
	    (new-matrix (truncated-matrix-multiply matrix old-matrix))) ;order matters
       (push old-matrix grf-matrix-stack)
       (setf grf-m0 (first new-matrix))
       (setf grf-m1 (second new-matrix))
       (setf grf-m2 (third new-matrix))
       (setf grf-m3 (fourth new-matrix))
       (setf grf-xoff (fifth new-matrix))
       (setf grf-yoff (sixth new-matrix))
       (setf grf-xp 0.0)
       (setf grf-yp 0.0))))
  nil)

(defun g-pop (score)
  (case (output-type score)
    (:postscript (format (output score) " grestore~%"))
    (:C (c-print " grestore"))
    (otherwise
     (let ((vals (pop grf-matrix-stack)))
       (setf grf-m0 (first vals))
       (setf grf-m1 (second vals))
       (setf grf-m2 (third vals))
       (setf grf-m3 (fourth vals))
       (setf grf-xoff (fifth vals))
       (setf grf-yoff (sixth vals))
       (setf grf-xp (seventh vals))
       (setf grf-yp (eighth vals)))))
  nil)


;;; -------------------------------- G-SET-LINE-WIDTH --------------------------------

(defun g-set-line-width (score xs)
  ;; set line-width for subsequent drawing (can be 0 = thinnest possible line)
  (case (output-type score)
    (:postscript (format (output score) " ~A setlinewidth~%" (not-rational xs)))
    (:C (setf (aref c-args 0) (make-short-float xs))
	(c-linewidth c-args))
    (:X (x-line-width (round xs)))
    #+(and mcl (not openmcl))
    (:QDDirect  
     (let ((ps (max 1 (ceiling xs))))
       (when (/= ps grf-width)
	 (d_pensize ps ps))))
    )
  (setf grf-width xs)
  nil)


;;; -------------------------------- G-SET-PATTERN --------------------------------
;;;
;;; if type is not nil or :outlined, we funcall the type field with the score as its argument.

(defun g-set-pattern (score type)
  (if (and type (not (eq type :outlined))) (funcall type score)))



;;; -------------------------------- G-SET-COLOR --------------------------------

(defun g-set-color (score rgb)
  (case (output-type score)
    (:postscript (format (output score) " ~{~,3F ~} setrgbcolor" rgb))
    (:C (c-print (format nil " ~{~,3F ~} setrgbcolor" rgb)))
    (otherwise
     (setf grf-color (copy-list rgb))
     (case (output-type score)
       (:X (x-color (round (* (first rgb) 256)) (round (* (second rgb) 256)) (round (* (third rgb) 256))))
       (otherwise 
	#+(and mcl (not openmcl))
	(ccl:set-fore-color 
	       (ccl:view-window *cmn-view*)
	       (ccl:make-color (round (* (first rgb) 65535)) (round (* (second rgb) 65535)) (round (* (third rgb) 65535))))
	))))
  nil)


;;; -------------------------------- G-SET-FONT --------------------------------

(defun g-set-font (score font size)
  ;; establish font as new default
  (case (output-type score)
    (:postscript (format (output score) " /~A findfont ~D scalefont setfont~%" font (round size)))
    (:C (c-print (format nil " /~A findfont ~D scalefont setfont" font (round size))))
    (otherwise
     (setf grf-font font)
     (setf grf-fontsize (round size))))
  nil)


;;; -------------------------------- G-MOVETO and G-RMOVETO
;;;
;;; Our underlying model here is postscript's notion of a current path which can later be filled and whatnot.
;;; In quickdraw and X this requires saving the points (and operations) until the actual draw or fill operation.

(defconstant MOVE-OP 0)
(defconstant LINE-OP 1)

(defun g-moveto (score xs ys)
  ;; move (without drawing) the current point to (xs ys)
  (case (output-type score)
    (:postscript (format (output score) " ~,2F ~,2F moveto~%" xs ys))
    (:C 
     (setf (aref c-args 0) (make-short-float xs))
     (setf (aref c-args 1) (make-short-float ys))
     (c-moveto c-args))
    (otherwise
     (setf grf-xp xs)
     (setf grf-yp ys)
     (push (list MOVE-OP xs ys) grf-xpoints)))
  nil)

(defun g-rmoveto (score dxs dys)
  ;; move the current point by (dxs dys)
  (case (output-type score)
    (:postscript (format (output score) " ~,2F ~,2F rmoveto~%" dxs dys))
    (:C
     (setf (aref c-args 0) (make-short-float dxs))
     (setf (aref c-args 1) (make-short-float dys))
     (c-rmoveto c-args))
    (otherwise
     (incf grf-xp dxs)
     (incf grf-yp dys)
     (push (list MOVE-OP grf-xp grf-yp) grf-xpoints)))
  nil)


;;; -------------------------------- LINETO and RLINETO

(defun g-lineto (score xs ys)
  ;; draw a line (hopefully in the current line-width) from the current point to (xs ys)
  ;; the current point in cmn is [(x score),(y score)]
  (case (output-type score)
    (:postscript (format (output score) " ~,2F ~,2F lineto~%" xs ys))
    (:C 
     (setf (aref c-args 0) (make-short-float xs))
     (setf (aref c-args 1) (make-short-float ys))
     (c-lineto c-args))
    (otherwise
     (let ((oxp grf-xp)
	   (oyp grf-yp))
       (setf grf-xp xs)
       (setf grf-yp ys)
       (push (list LINE-OP oxp oyp grf-xp grf-yp) grf-xpoints))))
  nil)

(defun g-rlineto (score dxs dys)
  ;; draw a line from the current point by (dxs dys)
  (case (output-type score)
    (:postscript (format (output score) " ~,2F ~,2F rlineto~%" dxs dys))
    (:C 	    
     (setf (aref c-args 0) (make-short-float dxs))
     (setf (aref c-args 1) (make-short-float dys))
     (c-rlineto c-args))
    (otherwise
     (let ((oxp grf-xp)
	   (oyp grf-yp))
       (incf grf-xp dxs)
       (incf grf-yp dys)
       (push (list LINE-OP oxp oyp grf-xp grf-yp) grf-xpoints))))
  nil)


;;; -------------------------------- CURVES and DASHED LINES BY HAND --------------------------------
;;;
;;; X (in some cases) and Quickdraw do not directly support Bezier curves.
;;; Quickdraw does not support dashed lines, and X only supports the simplest possible cases.
;;; All done here with moveto and lineto.

;;; here is the PostScript Bezier curve (curveto above) done by hand:
;;; (n is how accurate a curve we draw)

(defun output-bezier (score x0 y0 x1 y1 x2 y2 x3 y3 nn)
  ;; nn is the accuracy of the rendition
  (let* ((n (or nn (max 5 (round (size score) 4))))
	 (cx (* 3 (- x1 x0)))
	 (cy (* 3 (- y1 y0)))
	 (bx (- (* 3 (- x2 x1)) cx))
	 (by (- (* 3 (- y2 y1)) cy))
	 (ax (- x3 x0 cx bx))
	 (ay (- y3 y0 cy by))
	 (incr (/ 1.0 n)))
    (if (or (/= x0 (box-x0 score))
	    (/= y0 (box-y0 score)))
	(g-moveto score x0 y0))
    (loop for i from 0 to 1 by incr do
	  (g-lineto score 
		    (+ x0 (* i (+ cx (* i (+ bx (* i ax))))))
		    (+ y0 (* i (+ cy (* i (+ by (* i ay))))))))))

(defun g-dashed-line (score x y pattern rl)
  ;; can be horizontal or vertical (dashed bar line for example)
  (case (output-type score)
    (:postscript			
     (format (output score) " ~,2F ~,2F ~A [~{~A ~}] 0 setdash stroke~% [] 0 setdash~%" 
	     x y 
	     (if rl "rlineto" "lineto")
	     (map 'list #'not-rational pattern)))
    (:C     
     (setf (aref c-args 0) (make-short-float x))
     (setf (aref c-args 1) (make-short-float y))
     (if rl (c-rlineto c-args) (c-lineto c-args))
     (c-print (format nil " [~{~A ~}] 0 setdash stroke [] 0 setdash" (map 'list #'not-rational pattern))))
    (otherwise		; dashed lines in quickdraw are done by hand via line/move
     (let* ((dline (first pattern))
	    (dspace (second pattern))
	    (dlen (+ dspace dline)))
       (if (or (zerop dspace) (zerop dline))
	   (g-rlineto score x y)
	 (let* ((dx (if rl x (- x (scr-x score))))
		(dy (if rl y (- y (scr-y score))))
		(dashes (ceiling (sqrt (+ (* dx dx) (* dy dy))) dlen)))
	   (when (plusp dashes)
	     (let* ((ddx (if (zerop dy) dlen (/ dx dashes)))
		    (ddy (if (zerop dx) dlen (/ dy dashes)))
		    (dashx (* ddx (/ dline dlen)))
		    (dashy (* ddy (/ dline dlen)))
		    (spacex (- ddx dashx))
		    (spacey (- ddy dashy))
		    (lastdx (if (plusp dx) (- dx (* (1- dashes) dlen)) 0))
		    (lastdy (if (plusp dy) (- dy (* (1- dashes) dlen)) 0)))
	       (loop for i from 0 below (1- dashes) and 
                         x0 from (scr-x score) by ddx and 
                         y0 from (scr-y score) by ddy do
		 (g-rlineto score dashx dashy)
		 (g-rmoveto score spacex spacey))
	       (if (or (plusp lastdx) (plusp lastdy))
		   (g-rlineto score lastdx lastdy)))))))))
  nil)



;;; -------------------------------- BEZIER CURVES --------------------------------

(defun g-curveto (score x0 y0 x1 y1 x2 y2)
  ;; draw Bezier curve
  ;; Postscript uses a cubic Bezier curve, as opposed to the quadratic Quickdraw font-manager version
  (case (output-type score)
    (:postscript (format (output score) " ~,2F ~,2F ~,2F ~,2F ~,2F ~,2F curveto~%" x0 y0 x1 y1 x2 y2))
    (:C 	      
     (setf (aref c-args 0) (make-short-float x0))
     (setf (aref c-args 1) (make-short-float y0))
     (setf (aref c-args 2) (make-short-float x1))
     (setf (aref c-args 3) (make-short-float y1))
     (setf (aref c-args 4) (make-short-float x2))
     (setf (aref c-args 5) (make-short-float y2))
     (c-curveto c-args))
    (otherwise  (output-bezier score (scr-x score) (scr-y score) x0 y0 x1 y1 x2 y2 *grf-curve-accuracy*)))
  nil)



;;; -------------------------------- ARCS --------------------------------
;;;
;;; CMN only calls for an arc when it is assuming no current path is in progress
;;; (that is, from PostScript's point of view, there is always a preceding newpath).

(defvar linebuf (make-array 5 :element-type #-(or cmu sbcl) 'fixnum #+(or cmu sbcl) '(unsigned-byte 32)))
(defvar polybuf (make-array 1024 :element-type #-(or cmu sbcl) 'fixnum #+(or cmu sbcl) '(unsigned-byte 32)))
(defvar polybufsize 1024)

(defun g-arc (score x y r ang1 ang2 &optional fill)
  ;; arc of a circle centered at (x y) radius r in current line-width or (if fill) filled, partial arc if ang1 ang2
  ;; ang1=0 and ang2=360 is a full circle (very common special case)
  ;; in Postscript angle increases counterclockwise, 0=>x axis (3 o'clock), ang2 is from 0, not from ang1
  (case (output-type score)
    (:postscript (format (output score) " ~,3F ~,3F ~,3F ~D ~D newpath arc~A~%" x y r ang1 ang2 (if fill " fill" " stroke")))
    (:C (c-print (format nil " ~,3F ~,3F ~,3F ~D ~D newpath arc~A" x y r ang1 ang2 (if fill " fill" " stroke"))))
    (:X
     (setf grf-xpoints nil)            ;implicit newpath here
     (setf grf-xp x)
     (setf grf-yp y)
     (let ((x0 (position-x x y))
	   (y0 (position-y x y)))
       (setf (aref linebuf 0) x0)
       (setf (aref linebuf 1) y0)
       (setf (aref linebuf 2) (round (* grf-m0 r))) ;circles remain circular
       (setf (aref linebuf 3) (round ang1))
       (setf (aref linebuf 4) (round ang2))
       (x-circle (if fill 1 0) linebuf)))
    #+(and mcl (not openmcl))
    (:QDDirect  
     ;; here we need the implied circle's bounding box, with 0=>y axis upward (the negative y axis!!)
     ;; with the angle increasing clockwise
     (progn 
       (ccl::with-rectangle-arg
	 (rect (position-x (- x r) (+ y r))
	       (position-y (- x r) (+ y r))
	       (position-x (+ x r) (- y r))
	       (position-y (+ x r) (- y r)))
         (if (and (= ang1 0) (= ang2 360))
	     (if fill (d_PaintOval rect) (d_FrameOval rect))
           (if fill
	       (d_PaintArc rect (round (- 90 ang1)) (round (- ang1 ang2)))
             (d_FrameArc rect (round (- 90 ang1)) (round (- ang1 ang2))))))))
    )
  nil)


;;; -------------------------------- DRAW and FILL --------------------------------
;;;
;;; This is where all the actual work takes place

(defun g-draw (score &optional width)
  ;; if some object is awaiting a draw decision, send it out
  (case (output-type score)
    (:postscript 
     (if (and (numberp width) (/= width 0))
	 (format (output score) " ~D setlinewidth stroke 0 setlinewidth~%" width)
       (format (output score) " stroke~%")))
    (:C
     (if (and (numberp width) (/= width 0))
	 (progn
	   (setf (aref c-args 0) (float width))
	   (c-linewidthcleared c-args))
       (c-print " stroke")))
    (otherwise
     ;; draw and fill need to take the current path and either draw lines or fill polygons
     (let ((np (nreverse grf-xpoints)))
       (when width (g-set-line-width score width))
       (setf grf-xpoints nil)
       (loop for pt in np do
         ;; in this case (draw as opposed to fill) we need only grab linetos
	 ;; but line width matters (line is considered to be centered within the resultant polygon)
	  (if (= (first pt) LINE-OP)
	      (case (output-type score)
		(:X
		 (setf (aref linebuf 0) (position-x (second pt) (third pt)))
		 (setf (aref linebuf 1) (position-y (second pt) (third pt)))
		 (setf (aref linebuf 2) (position-x (fourth pt) (fifth pt)))
		 (setf (aref linebuf 3) (position-y (fourth pt) (fifth pt)))
		 (x-lineto linebuf))
		#+(and mcl (not openmcl))
		(:QDDirect
		 (d_lineto (position-x (fourth pt) (fifth pt)) (position-y (fourth pt) (fifth pt))))
		)
	    #+(and mcl (not openmcl))
	    (case (output-type score)
	      ;; must be a moveto -- only affects quickdraw
	      (:QDDirect
	       (d_moveto (position-x (second pt) (third pt)) (position-y (second pt) (third pt)))))
	    )))))
  nil)

(defun g-fill (score cmd even-odd)
  ;; here we need to grab all the points as a possibly self-intersecting polygon
  (case (output-type score)
    (:postscript (format (output score) " ~A~%" cmd))
    (:C (c-print cmd))
    (:X
     (let* ((np (nreverse grf-xpoints))
	    (len (length np)))
       (setf grf-xpoints nil)
       (when (> (* 2 len) polybufsize)
	 (setf polybufsize (+ (* 2 len) 512))
	 (setf polybuf (make-array polybufsize :element-type #-(or cmu sbcl) 'fixnum #+(or cmu sbcl) '(unsigned-byte 32))))
       (loop for pt in np and i from 0 by 2 do
	 (if (= (first pt) MOVE-OP)
	     (progn
	       (setf (aref polybuf i) (position-x (second pt) (third pt)))
	       (setf (aref polybuf (+ i 1)) (position-y (second pt) (third pt))))
	   (progn
	     (setf (aref polybuf i) (position-x (fourth pt) (fifth pt)))
	     (setf (aref polybuf (+ i 1)) (position-y (fourth pt) (fifth pt))))))
       (x-polygon 1 (if even-odd 1 0) len polybuf)))
    #+(and mcl (not openmcl))
    (:QDDirect
     (let* ((np (nreverse grf-xpoints)))
       (setf grf-xpoints nil)
       (setf (ccl::view-get *cmn-view* 'my-poly) (d_OpenPoly))
       (loop for pt in np do
	 (if (= (first pt) MOVE-OP)
	     (d_moveto (position-x (second pt) (third pt)) (position-y (second pt) (third pt)))
	   (d_lineto (position-x (fourth pt) (fifth pt)) (position-y (fourth pt) (fifth pt)))))
       (let ((new-poly (ccl::view-get *cmn-view* 'my-poly)))
         (d_ClosePoly)
         (d_PaintPoly new-poly)
         (d_KillPoly new-poly))))
    )
  nil)


;;; -------------------------------- TEXT --------------------------------

(defun g-text (score txt &optional outlined)
  ;; in PS, assume font is set up
  ;; in others, make sure current font is active
  (case (output-type score)
    (:postscript (format (output score) " (~A) ~A~%" (ps-letters txt) (if outlined "false charpath stroke" "show")))
    (:C (c-print (format nil " (~A) ~A" (ps-letters txt) (if outlined "false charpath stroke" "show"))))
    (:X 
     (let ((x0 (position-x grf-xp grf-yp))
	   (y0 (position-y grf-xp grf-yp)))
       (setf grf-xpoints nil)
       (setf (aref linebuf 0) x0)
       (setf (aref linebuf 1) y0)
       (setf grf-active-font grf-font)
       (setf grf-active-fontsize grf-fontsize)
       (x-text (round grf-fontsize) linebuf (Ps-to-X-font-name (or grf-font (normal-font)) (round grf-fontsize)) txt)))
    #+(and mcl (not openmcl))
    (otherwise
     (when (or (not (eq grf-font grf-active-font))
	       (/= grf-fontsize grf-active-fontsize))
       (let* ((qd-font-data (ps-to-qd-font-name grf-font))
	      (qd-font (first qd-font-data))
	      (qd-type (encode-text-face (second qd-font-data)))
	      (qd-font-number (ash (ccl:font-codes (list qd-font grf-fontsize (second qd-font-data))) -16)))
	 (d_TextSize grf-fontsize)
	 (d_TextFace (+ qd-type (if outlined 8 0)))
	 (d_TextFont qd-font-number)))
     (setf grf-active-font grf-font)
     (setf grf-active-fontsize grf-fontsize)
     (d_moveto (position-x grf-xp grf-yp) (position-y grf-xp grf-yp))
     (d_drawstring txt)))
  nil)


(defun g-mustext (score txt dx dy)
  (when (output score)
    (moveto score dx dy)
    (g-set-font score (music-font score) (size score))
    (case (output-type score)
      (:postscript (format (output score) " (\\~O) show~%" txt))
      (:C (c-print (format nil " (\\~O) show" txt)))
      (:X 
       (let ((x0 (position-x grf-xp grf-yp))
	     (y0 (position-y grf-xp grf-yp)))
	 (setf grf-xpoints nil)
	 (setf (aref linebuf 0) x0)
	 (setf (aref linebuf 1) y0)
	 (setf grf-active-font grf-font)
	 (setf grf-active-fontsize grf-fontsize)
	 (x-text (round grf-fontsize) linebuf (Ps-to-X-font-name (or grf-font (normal-font)) (round grf-fontsize)) txt)))
      #+(and mcl (not openmcl))
      (otherwise
       (when (or (not (eq grf-font grf-active-font))
		 (/= grf-fontsize grf-active-fontsize))
	 (let* ((qd-font-data (ps-to-qd-font-name grf-font))
		(qd-font (first qd-font-data))
		(qd-type (encode-text-face (second qd-font-data)))
		(qd-font-number (ash (ccl:font-codes (list qd-font grf-fontsize (second qd-font-data))) -16)))
	   (d_TextSize grf-fontsize)
	   (d_TextFace qd-type)
	   (d_TextFont qd-font-number)))
       (setf grf-active-font grf-font)
       (setf grf-active-fontsize grf-fontsize)
       (d_moveto (position-x grf-xp grf-yp) (position-y grf-xp grf-yp))
       (d_drawchar txt))))
  nil)



  
;;; -------------------------------- OPEN and CLOSE --------------------------------

(defun x-close ()
  (when xcmnw-open
    (cmn-pclose)
    (setf xcmnw-open nil)))
    
(defun g-header (score title user-name)
  (case (output-type score)
    (:postscript 
     (progn 
       (format (output score) "%!PS-Adobe-2.0 EPSF-2.0~%") ;this is what our Next's seem to have -- Adobe recommends 3.0
       ;; exact caps matter here --if you had %!Ps-Adobe, WriteNow would throw up its hands and refuse to paste it in.
       (if title (format (output score) "%%Title: ~A~%" title))
       (if user-name (format (output score) "%%Creator: ~A~%" user-name))
       (format (output score) "%%CreationDate: ~A~%" (creation-date))
       (if (bounded score)
	   (progn
	     (setf (box-x0 score) (first (bounded score)))
	     (setf (box-y0 score) (second (bounded score)))
	     (setf (box-x1 score) (third (bounded score)))
	     (setf (box-y1 score) (fourth (bounded score)))
	     (format (output score) "%%BoundingBox:~{ ~D~}~%" (map 'list #'ceiling (bounds score))))
	 (format (output score) "%%BoundingBox:(atend)~%"))
       (format (output score) "%%EndComments~%")
       (if (prologue score) (loop for head in (reverse (prologue score)) do (funcall head score)))
       (format (output score) "%%EndProlog~%")
       (format (output score) "%%Page: 1 1~%")
       (if (page-color score) (format (output score) "clippath ~{~,3F ~} setrgbcolor fill " (page-color score)))
       (format (output score) "~{~,3F ~} setrgbcolor" (or (color score) '(0 0 0)))))
    (:C
     (c-print (format nil "%!PS-Adobe-2.0 EPSF-2.0"))
     (if title (c-print (format nil "%%Title: ~A" title)))
     (if user-name (c-print (format nil "%%Creator: ~A" user-name)))
     (c-print (format nil "%%CreationDate: ~A" (creation-date)))
     (if (bounded score)
	 (progn
	   (setf (box-x0 score) (first (bounded score)))
	   (setf (box-y0 score) (second (bounded score)))
	   (setf (box-x1 score) (third (bounded score)))
	   (setf (box-y1 score) (fourth (bounded score)))
	   (c-print (format nil "%%BoundingBox:~{ ~D~}" (map 'list #'ceiling (bounds score)))))
       (c-print (format nil "%%BoundingBox:(atend)")))
     (c-print (format nil "%%EndComments"))
     (if (prologue score) (loop for head in (reverse (prologue score)) do (funcall head score)))
     (c-print (format nil "%%EndProlog"))
     (c-print "%%Page: 1 1")
     (if (page-color score) (c-print (format nil "clippath ~{~,3F ~} setrgbcolor fill" (page-color score))))
     (c-print (format nil "~{~,3F ~} setrgbcolor" (or (color score) '(0 0 0)))))
    (:X
     (g-clear-context score)
     (when (not xcmnw-open)
       (cmn-popen (concatenate 'string *cmn-binary-directory* "xcmnw"))
       (setf xcmnw-open t))
     (x-clear-window))
    #+(and mcl (not openmcl))
    (:QDDirect 
     (g-clear-context score))
    )
  ;;(if (color score) (g-set-color score (color score)))
  nil)

(defun g-footer (score)
  (case (output-type score)
    (:postscript (format (output score) " showpage~%%%Trailer~%"))
    (:C (c-print (format nil " showpage~%%%Trailer")))
    )
  nil)

(defun g-bbox (score)
  (case (output-type score)
    (:postscript (format (output score) "%%BoundingBox:~{ ~D~}~%" (map 'list #'(lambda (n) (round (1+ (* (scr-size score) n)))) (bounds score))))
    (:C (c-print (format nil "%%BoundingBox:~{ ~D~}~%" (map 'list #'(lambda (n) (round (1+ (* (scr-size score) n)))) (bounds score)))))
    ;(:postscript (format (output score) "%%BoundingBox:~{ ~D~}~%" (map 'list #'(lambda (n) (round (1+ n))) (bounds score))))
    ;(:C (c-print (format nil "%%BoundingBox:~{ ~D~}~%" (map 'list #'(lambda (n) (round (1+ n))) (bounds score)))))
    )
  nil)

(defun g-send (score cmd)
  (case (output-type score)
    (:postscript (format (output score) " ~A~%" cmd))
    (:C (c-print (format nil " ~A" cmd)))
    )
  nil)

(defun g-page-number (score page eject)
  (case (output-type score)
    (:postscript
     (when eject (format (output score) " showpage~%"))
     (format (output score) "%%Page: ~D ~D~%" page page))
    (:C
     (when eject (c-print "showpage"))
     (c-print (format nil "%%Page: ~D ~D~%" page page)))
    )
  nil)



#+(and excl cltl2) (eval-when (compile load eval) (excl:advise excl:exit :before cmn-exit-advice nil (x-close)))
    
#+cmu (defun bye () (x-close) (extensions::quit)) ;who can remember which lisp uses which!?!?!?
#+cmu (defun exit () (x-close) (extensions::quit))
#+sbcl (defun bye () (sb-ext:quit))
#+sbcl (defun exit () (sb-ext:quit))
#+(and clisp (not ansi-cl) (not have-quit)) (defun quit () (lisp:bye))
#+(and clisp ansi-cl (not have-quit)) (defun quit () (bye))
#+(and clisp (not ansi-cl) (not have-quit)) (defun exit () (lisp:bye))
#+(and clisp ansi-cl (not have-quit)) (defun exit () (bye))
#+excl (defun bye () (excl:exit))
#+excl (defun quit () (excl:exit))
#+openmcl (defun bye () (ccl::quit))
#+openmcl (defun exit () (ccl::quit))
;;; nothing is more annoying than a program that will not exit!
